/*
 *  Header file for info interface
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

/* buffer for informations */

struct snd_info_buffer {
  char *buffer;                 /* pointer to begin of buffer */
  char *curr;                   /* current position in buffer */
  unsigned long size;           /* current size */
  unsigned long len;            /* total length of buffer */
  int stop;                     /* stop flag */
  int error;                    /* error code */
};

typedef struct snd_info_buffer snd_info_buffer_t;

struct snd_info_entry {
  const char *name;
  mode_t mode;
  unsigned long read_size;
  unsigned long write_size;
  void (*read)( snd_info_buffer_t *buffer, void *private_data );
  void (*write)( snd_info_buffer_t *buffer, void *private_data );
  void (*use_inc)( snd_card_t *card );
  void (*use_dec)( snd_card_t *card );
  void *private_data;
  snd_card_t *card;
  struct proc_dir_entry *p;
  struct snd_info_entry *next;
};

extern int snd_iprintf( snd_info_buffer_t *buffer, char *fmt, ... );
extern int snd_info_init( void );
extern int snd_info_done( void ); 
extern int snd_info_check_reserved_words( const char *str );
extern snd_info_entry_t *snd_info_create_entry( snd_card_t *card, const char *name );
extern void snd_info_free_entry( snd_info_entry_t *entry );
extern int snd_info_card_register( snd_card_t *card );
extern int snd_info_card_unregister( snd_card_t *card );
extern int snd_info_register( snd_info_entry_t *entry );
extern int snd_info_unregister( snd_info_entry_t *entry );
extern int snd_info_minor_register( void );
